using System;
using System.IO;
using System.Collections;
using System.Runtime.Serialization.Formatters.Binary;

using jade.core;
using jade.lang.acl;

using log4net;

using Mailabs.Reasoning;

namespace Mailabs.Tutorial.MAS
{
    public class RequestProcessor
    {
        private static readonly ILog logger = LogManager.GetLogger(typeof(RequestProcessor));

        private static Engine engine = new Engine();

        private static Hashtable agents = new Hashtable();

        internal static void ProcessRequest(CommAgent agent, ReasoningEntity request, bool isInitial)
        {
            if (isInitial)
            {
                lock (typeof(RequestProcessor))
                {
                    agents[request] = agent;
                }
            }

            ReasoningEntity profile = agent.profile;
            engine.Reset();
            engine.LoadGraph(new ReasoningEntity[] { profile, request });
            double relevancy = engine.Relevancy(profile, request);
            logger.Debug("Compared profile to request, relevancy is: " + relevancy);
            if (relevancy >= Consts.ACQUAINTANCE_RELEVANCY_THRESHOLD)
            {
                logger.Info(">>> Success, profile of " + profile.Root.RelativeID + " and current request are in good fit!");
                if (isInitial)
                {
                    logger.Debug(">>> The request placed in the agent's folder is suitable for the agent itself");
                }
                else
                {
                    ACLMessage response = new ACLMessage(ACLMessage.PROPAGATE);
                    Agent requestOriginator = (Agent)agents[request];
                    response.addReceiver(requestOriginator.getAID());
                    logger.Debug(">>> Sending response to the request originator: " + requestOriginator.getLocalName());
                    response.setByteSequenceContent(GetBytesFromObject(profile));
                    agent.send(response);
                    logger.Debug("Response sent.");
                }
            }

            // TODO: in future versions of ORe there will be no need in relevance normalization
            double newWeight = relevancy > 0 ? request.Weight * 0.8D * relevancy / 0.0555555555555556D : request.Weight * 0.2D;

            logger.Debug("New request weight is " + newWeight);

            if (newWeight >= Consts.RESEND_WEIGHT_THRESHOLD)
            {
                request.Weight = newWeight;
                if (agent.friends.Count > 0)
                {
                    ACLMessage message = new ACLMessage(ACLMessage.INFORM);
                    for (IEnumerator e = agent.friends.GetEnumerator(); e.MoveNext(); )
                    {
                        string friendName = (string)e.Current;
                        message.addReceiver(new AID(friendName, AID.ISLOCALNAME));
                        logger.Debug("Sending message to " + friendName);
                    }

                    message.setByteSequenceContent(GetBytesFromObject(request));
                    agent.send(message);
                    logger.Debug("Message sent.");
                }
                else
                {
                    logger.Debug("No friends to resend message to, stop processing");
                    lock (typeof(RequestProcessor))
                    {
                        agents.Remove(request);
                    }
                }
            }
            else
            {
                logger.Debug("Request weight too low, stop processing");
                lock (typeof(RequestProcessor))
                {
                    agents.Remove(request);
                }
            }
        }

        internal static void ProcessRequest(CommAgent agent, sbyte[] image)
        {
            ProcessRequest(agent, GetObjectFromBytes(image), false);
        }

        internal static sbyte[] GetBytesFromObject(ReasoningEntity request)
        {
            MemoryStream stream = new MemoryStream();
            BinaryFormatter formatter = new BinaryFormatter();
            formatter.Serialize(stream, request);
            sbyte[] result = new sbyte[stream.GetBuffer().Length];
            for (int i = 0; i < result.Length; i++)
            {
                result[i] = (sbyte)stream.GetBuffer()[i];
            }
            return result;
        }

        internal static ReasoningEntity GetObjectFromBytes(sbyte[] image)
        {
            byte[] newImage = new byte[image.Length];
            for (int i = 0; i < newImage.Length; i++)
            {
                newImage[i] = (byte)image[i];
            }
            MemoryStream stream = new MemoryStream(newImage);
            BinaryFormatter formatter = new BinaryFormatter();
            ReasoningEntity result = (ReasoningEntity)formatter.Deserialize(stream);

            return result;
        }
    }
}
