/*

  ORe - OWL ontology relevance library for Microsoft .NET
  (http://ore.mailabs.ru)

  Copyright (c) 2005, 2006 Sergey V. Mikhanov (MAILabs)

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  Version: $Id: $
  
*/

using System;
using System.IO;
using System.Net;
using System.Collections;
using System.Globalization;

using log4net;

namespace Mailabs.Reasoning
{
    public class CacheService
    {
        private static readonly ILog logger = LogManager.GetLogger(typeof(CacheService));

        private CacheService() {}

        static CacheService() 
        {
            if (!Directory.Exists(Consts.LOCAL_ONTOLOGIES_CACHE)) 
            {
                Directory.CreateDirectory(Consts.LOCAL_ONTOLOGIES_CACHE);
            }
        }

        private static Hashtable ontologies = new Hashtable();

        public static Ontology CacheFor(string uri) 
        {
            return CacheFor(uri, LoadCallback.NULL);
        }

        public static Ontology CacheFor(string uri, LoadCallback callback) 
        {
            lock (typeof(CacheService))
            {
                Ontology result;

                logger.Debug("Cache initialized");
                callback.OnCacheInit();

                if ((result = (Ontology)ontologies[uri]) != null)
                {
                    logger.Debug("Found cached version of the ontology " + uri + ", checking whether it is obsolete");
                    try
                    {
                        DateTime lastModified = LastModified(uri);
                        if (lastModified > result.LastModified)
                        {
                            logger.Debug("Ontology " + uri + " needs renovation");
                            Uri src = new Uri(uri);

                            if (!src.IsFile)
                            {
                                logger.Debug("Creating locally cached version of the remote ontology: " + src);
                                callback.OnCacheCreate();
                                CreateLocalCache(src, new Uri(LocalPath(uri)));
                            }

                            logger.Debug("Creating in-memory copy from the local copy (or just source file, if using a file)");
                            result = new Ontology(LocalPath(uri), callback);
                            ontologies.Remove(uri);
                            ontologies.Add(uri, result);
                            return result;
                        }
                        else
                        {
                            logger.Debug("Ontology " + uri + " do not need renovation, returning in-memory copy");
                            return result;
                        }
                    }
                    catch (WebException)
                    {
                        logger.Debug("Ontology " + uri + " do not need renovation, returning in-memory copy");
                        return result;
                    }
                }
                else
                {
                    try
                    {
                        Uri src = new Uri(uri);
                        Uri dest = new Uri(LocalPath(uri));

                        if (!src.IsFile)
                        {
                            if (File.Exists(dest.LocalPath))
                            {
                                DateTime fileTime = File.GetLastWriteTime(dest.LocalPath);
                                DateTime webTime = LastModified(src.AbsoluteUri);

                                if (webTime > fileTime)
                                {
                                    logger.Debug("Creating cached version of the ontology: " + src);
                                    callback.OnCacheCreate();
                                    CreateLocalCache(src, dest);
                                }
                            }
                            else
                            {
                                logger.Debug("Creating cached version of the ontology: " + src);
                                callback.OnCacheCreate();
                                CreateLocalCache(src, dest);
                            }
                        }

                        result = new Ontology(LocalPath(uri), callback);
                        ontologies.Add(uri, result);
                        return result;
                    }
                    catch (Exception e)
                    {
                        logger.Error("Error while caching ontology: " + uri, e);
                        callback.OnError(e);
                        result = LocalCache(uri, callback);

                        if (result != null)
                            ontologies.Add(uri, result);
                        return result;
                    }
                }
            }
        }
        
        private static DateTime LastModified(string uri) 
        {
            DateTime lastModified;
            Uri ontologyUri = new Uri(uri);
            if (ontologyUri.IsFile) 
            {
                lastModified = File.GetLastWriteTime(ontologyUri.LocalPath);
            } 
            else 
            {
                using (WebClient client = new WebClient()) 
                {
                    Stream stream = client.OpenRead(ontologyUri.AbsoluteUri);

                    lastModified = DateTime.MinValue;

                    if (client.ResponseHeaders["Last-Modified"] != null && client.ResponseHeaders["Last-Modified"] != "") 
                    {
                        lastModified = DateTime.Parse(client.ResponseHeaders["Last-Modified"], new CultureInfo("en-US").DateTimeFormat);
                    } 
                    else 
                    {
                        lastModified = DateTime.Now;
                    }
                }
            }

            return lastModified;
        }

        private static void CreateLocalCache(Uri src, Uri dest) 
        {
            using (WebClient client = new WebClient()) 
            {
                client.DownloadFile(src.AbsoluteUri, dest.LocalPath);
            }
        }

        private static Ontology LocalCache(string uri, LoadCallback callback)
        {
            uri = LocalPath(uri);

            if (File.Exists(uri))
                return new Ontology(uri, callback);
            else
                return null;
        }

        private static string LocalPath(string remoteUri) 
        {
            string uri = remoteUri;
            if (new Uri(uri).IsFile)
                return uri;

            uri = uri.Replace('/', '_');
            uri = uri.Replace(':', '_');
            uri = uri.Replace('?', '_');
            uri = uri.Replace('.', '_');
            uri = uri.Replace('|', '_');

            uri = Consts.LOCAL_ONTOLOGIES_CACHE + uri;

            return uri;
        }
    }
}
