/*

  ORe - OWL ontology relevance library for Microsoft .NET
  (http://ore.mailabs.ru)

  Copyright (c) 2005, 2006 Sergey V. Mikhanov (MAILabs)

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  Version: $Id: $
  
*/

using System;
using System.Collections;

using log4net;

namespace Mailabs.Reasoning
{
    /// <summary>
    ///   .       OWL-.
    /// </summary>
    public class Engine
    {
        private static log4net.ILog logger;

        static Engine()
        {
            logger = LogManager.GetLogger(typeof(Engine));
        }

        private OwlGraph graph;

        public OwlGraph Graph 
        {
            get 
            {
                return graph;
            }
        }

        public Engine()
        {
            logger.Debug("Creating Engine");
            Reset();
        }

        /// <summary>
        ///    ( ).
        /// </summary>
        public void Reset()
        {
            graph = new OwlGraph();
        }

        /// <summary>
        ///         ,   ,
        ///   ,    .
        ///      !
        /// </summary>
        public void LoadGraph(ReasoningEntity[] entities)
        {
            graph = new OwlGraph();
            for (int i = 0; i < entities.Length; i++) 
            {
                Merge(entities[i]);
            }
        }

        /// <summary>
        ///       ,
        ///   ,    .
        /// </summary>
        public void Merge(ReasoningEntity entity) 
        {
            if (entity == null)
            {
                logger.Debug("Attemted to merge null entity, ignoring");
                return;
            }

            OwlClass node = entity.Root;

            logger.Debug("Merging " + node.RelativeID);
                
            if (!graph.Classes.Contains(node.ID)) 
            {
                if (node.IsRemote) 
                {
                    node = Fetch(node);
                    if (node == null) 
                    {
                        // for some reasons ontology haven't been received,
                        // so there's nothing more to do here
                        return;
                    }
                }
                graph.Classes.Add(node.ID, node);
            }

            CopyDictionary(node.Parents, graph.Classes);
            CopyDictionary(node.Equivalents, graph.Classes);
            CopyDictionary(node.Complements, graph.Classes);

            for (int j = 0; j < node.Restrictions.Count; j++) 
            {
                OwlRestriction rest = node.Restrictions[j];

                if (rest.Descriminator == OwlRestriction.OWL_CLASS_RESTRICTION)
                {
                    if (!graph.Classes.Contains(rest.GetChildClass().ID))
                    {
                        if (rest.GetChildClass().IsRemote)
                        {
                            OwlClass child = Fetch(rest.GetChildClass());
                            if (child == null)
                            {
                                // for some reasons ontology haven't been received,
                                // so there's nothing more to do here
                                return;
                            }
                            graph.Classes.Add(child.ID, child);
                        }
                        else
                        {
                            graph.Classes.Add(rest.GetChildClass().ID, rest.GetChildClass());
                        }
                    }
                }

                if (!graph.Roles.Contains(rest.Role.ID)) 
                {
                    if (rest.Role.IsRemote) 
                    {
                        OwlRole role = Fetch(rest.Role);
                        if (role == null) 
                        {
                            // for some reasons ontology haven't been received,
                            // so there's nothing more to de here
                            return;
                        }
                        graph.Roles.Add(role.ID, role);
                    } 
                    else 
                    {
                        graph.Roles.Add(rest.Role.ID, rest.Role);
                    }
                }
            }
        }

        private void CopyDictionary(IDictionary src, IDictionary dest)
        {
            for (IDictionaryEnumerator e = src.GetEnumerator(); e.MoveNext(); ) 
            {
                if (!dest.Contains(e.Key)) 
                {
                    dest.Add(e.Key, e.Value);
                }
            }
        }

        private OwlClass Fetch(OwlClass node) 
        {
            string remoteOntologyUri = node.ID.Substring(0, node.ID.IndexOf('#'));
            Ontology remoteOntology = new Ontology(remoteOntologyUri);
            return remoteOntology.Classes[node.ID];
        }

        private OwlRole Fetch(OwlRole role) 
        {
            string remoteOntologyUri = role.ID.Substring(0, role.ID.IndexOf('#'));
            Ontology remoteOntology = new Ontology(remoteOntologyUri);
            return remoteOntology.Roles[role.ID];
        }

        /// <summary>
        ///        .
        ///             ID  ReasoningException.
        ///  !   --   (, ),  --
        /// ,     (, ).
        /// </summary>
        public double Relevancy(ReasoningEntity ent1, ReasoningEntity ent2) 
        {
            OwlClass node1 = ent1.Root;
            OwlClass node2 = ent2.Root;

            if (graph.Classes[node1.ID] == null || graph.Classes[node2.ID] == null)
                throw new ReasoningException("Unable to start reasoning, classes '" + node1.ID + "' and '" + node2.ID + "' are not loaded properly!");

            OwlClassDictionary allParents1 = new OwlClassDictionary();
            // O(total number of parents of node1)
            FillParents(node1, allParents1);
            OwlClassDictionary allParents2 = new OwlClassDictionary();
            // O(total number of parents of node2)
            FillParents(node2, allParents2);
            OwlClassDictionary allRoles1 = new OwlClassDictionary();
            // O(number of role of node1)
            FillRoles(node1, allRoles1);
            OwlClassDictionary allRoles2 = new OwlClassDictionary();
            // O(number of role of node2)
            FillRoles(node2, allRoles2);

            
            double pp = GroupRelevancy(allParents1, allParents2);
            double pe = GroupRelevancy(allParents1, node2.Equivalents);
            double pc = GroupRelevancy(allParents1, node2.Complements);
            double pr = GroupRelevancy(allParents1, allRoles2);
            double ep = GroupRelevancy(node1.Equivalents, allParents2);
            double cp = GroupRelevancy(node1.Complements, allParents2);
            double rp = GroupRelevancy(allRoles1, allParents2);
            double ee = GroupRelevancy(node1.Equivalents, node2.Equivalents);
            double rr = GroupRelevancy(allRoles1, allRoles2);

            return (Weights.PP * pp
                + Weights.PE * (pe + ep)
                + Weights.PC * (pc + cp)
                + Weights.PR * (pr + rp)
                + Weights.EE * ee
                + Weights.RR * rr) / 9D;
        }

        // this method executes in O(dic1.Count * dic2.Count * mid(#par node from dics))
        private double GroupRelevancy(OwlClassDictionary dic1, OwlClassDictionary dic2) 
        {
            if (dic1.Count == 0 || dic2.Count == 0)
                return 0;

            double result = 0;
            for (IDictionaryEnumerator e = dic1.GetEnumerator(); e.MoveNext(); ) 
            {
                for (IDictionaryEnumerator g = dic2.GetEnumerator(); g.MoveNext(); ) 
                {
                    result += NodeRelevancy((OwlClass)e.Value, (OwlClass)g.Value);
                }
            }

            return result / (double)(dic1.Count * dic2.Count);
        }

        // O(#par node1 + #par node2)
        private double NodeRelevancy(OwlClass node1, OwlClass node2) 
        {
            if (node1.Equals(node2))
                return 1;

            Hashtable lengths1 = new Hashtable();
            lengths1.Add(node1.ID, 0);

            GatherParents(node1, lengths1, 1);
            
            Hashtable lengths2 = new Hashtable();
            if (lengths1.ContainsKey(node2.ID)) 
            {
                lengths2.Add(node2.ID, 0);
            }

            CompareParents(node2, lengths1, lengths2, 1);

            double maxRel = 0;
            for (IDictionaryEnumerator e = lengths2.GetEnumerator(); e.MoveNext(); ) 
            {
                int d_sp = (int)e.Value;
                int d_c = (int)lengths1[e.Key];
                double rel = DistanceRelevancy(d_sp, d_c);
                if (rel > maxRel) 
                    maxRel = rel;
            }

            return maxRel;
        }

        // assuming that Hastable.Item() executes in O(1), this method executes in
        // O(total number of parents of node)
        private void GatherParents(OwlClass node, Hashtable lengths, int level) 
        {
            for (IDictionaryEnumerator e = node.Parents.GetEnumerator(); e.MoveNext(); ) 
            {
                OwlClass parent = (OwlClass)e.Entry.Value;
                if (!lengths.ContainsKey(parent.ID)) 
                {
                    lengths.Add(parent.ID, level);
                } 
                else if ((int)lengths[parent.ID] > level) 
                {
                    lengths.Remove(parent.ID);
                    lengths.Add(parent.ID, level);
                }
                GatherParents(parent, lengths, level + 1);
            }
        }

        // assuming that Hastable.Item() executes in O(1), this method executes in
        // O(total number of parents of node)
        private void CompareParents(OwlClass node, Hashtable lengths1, Hashtable lengths2, int level) 
        {
            for (IDictionaryEnumerator e = node.Parents.GetEnumerator(); e.MoveNext(); ) 
            {
                OwlClass parent = (OwlClass)e.Entry.Value;
                if (lengths1.ContainsKey(parent.ID)) 
                {
                    if (!lengths2.ContainsKey(parent.ID)) 
                    {
                        lengths2.Add(parent.ID, level);
                    } 
                    else 
                    {
                        if ((int)lengths2[parent.ID] > level)
                        {
                            lengths2.Remove(parent.ID);
                            lengths2.Add(parent.ID, level);
                        }
                    }
                }

                CompareParents(parent, lengths1, lengths2, level + 1);
            }
        }

        private double DistanceRelevancy(int d_sp, int d_c) 
        {
            if (d_sp == 0)
                return 1;

            double dc = (double)d_c;
            double dsp = (double)d_sp;

            return 1 / (dc * dc + dsp * dsp);
        }

        private void FillParents(OwlClass cls, OwlClassDictionary dic) 
        {
            for (IDictionaryEnumerator e = cls.Parents.GetEnumerator(); e.MoveNext(); ) 
            {
                if (!dic.Contains((string)e.Key)) 
                {
                    dic.Add((string)e.Key, (OwlClass)e.Value);
                    FillParents((OwlClass)e.Value, dic);
                }
            }
        }

        private void FillRoles(OwlClass cls, OwlClassDictionary dic) 
        {
            for (IEnumerator e = cls.Restrictions.GetEnumerator(); e.MoveNext(); ) 
            {
                OwlRestriction rest = (OwlRestriction)e.Current;
                if (rest.Descriminator == OwlRestriction.OWL_CLASS_RESTRICTION)
                {
                    if (!rest.Negated)
                    {
                        if (!dic.Contains(rest.GetChildClass().ID))
                        {
                            dic.Add(rest.GetChildClass().ID, rest.GetChildClass());
                        }
                    }
                }
            }
        }
    }

    internal class Weights 
    {
        public const double PP = 1;
        public const double PE = 1;
        public const double PC = 0;
        public const double PR = 1;
        public const double EE = 1;
        public const double RR = 1;
    }
}
