/*

  ORe - OWL ontology relevance library for Microsoft .NET
  (http://ore.mailabs.ru)

  Copyright (c) 2005, 2006 Sergey V. Mikhanov (MAILabs)

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  Version: $Id: $
  
*/

using System;
using System.IO;
using System.Net;
using System.Text;
using System.Globalization;
using System.Collections;

using Drive.Rdf;

namespace Mailabs.Reasoning
{
    /// <summary>
    /// ,   (  AbstractOwlGraph).
    ///       ,   .
    /// </summary>
    public class Ontology : AbstractOwlGraph
    {
        /// <summary>
        /// URI .      ,    
        ///  ,    .  99.9%   ,    0.1%
        ///     .
        /// </summary>
        internal string _Uri;

        public string Uri 
        {
            get 
            {
                return _Uri;
            }
        }

        private DateTime _lastModified;

        /// <summary>
        ///     .       ,
        /// . .       --      DateTime.Now
        /// </summary>
        public DateTime LastModified 
        {
            get 
            {
                return _lastModified;
            }
        }

        /// <summary>
        ///   .
        /// </summary>
        public readonly OwlClassDictionary Classes;

        /// <summary>
        ///   .
        /// </summary>
        public readonly OwlRoleDictionary Roles;

        /// <summary>
        ///      OWL-,    URI.
        ///  callback-     .
        /// </summary>
        internal Ontology(string uri, LoadCallback callback)
        {
            Classes = new OwlClassDictionary();
            Roles = new OwlRoleDictionary();

            Uri ontologyUri = new Uri(uri);
            if (ontologyUri.IsFile) 
            {
                _lastModified = File.GetLastWriteTime(ontologyUri.LocalPath);
            } 
            else 
            {
                WebClient client = new WebClient();
                Stream stream = client.OpenRead(ontologyUri.AbsoluteUri);

                _lastModified = DateTime.MinValue;

                if (client.ResponseHeaders["Last-Modified"] != null && client.ResponseHeaders["Last-Modified"] != "") 
                {
                    _lastModified = DateTime.Parse(client.ResponseHeaders["Last-Modified"], new CultureInfo("en-US").DateTimeFormat);
                } 
                else 
                {
                    _lastModified = DateTime.Now;
                }

                stream.Close();
            }

            RdfHelper.Load(this, uri, callback);
        }

        /// <summary>
        ///      OWL-,    URI.
        /// </summary>
        internal Ontology(string uri) : this(uri, LoadCallback.NULL) {}

        protected string Crop(string fullID)
        {
            if (fullID.StartsWith(Uri))
            {
                return fullID.Substring(Uri.Length);
            }
            else
            {
                return fullID;
            }
        }

        public override string ToString()
        {
            StringBuilder builder = new StringBuilder();

            builder.Append("=== Ontology: " + Uri + " >>> ===\n");

            builder.Append("== Classes: " + Classes.Count + "\n");
            for (IDictionaryEnumerator e = Classes.GetEnumerator(); e.MoveNext(); )
            {
                OwlClass current = (OwlClass)e.Value;
                if (current.Parents.Count == 0)
                {
                    builder.Append(Crop(current.ID) + "\n");
                    AppendChildren(builder, Classes, current, 1);
                } else if (HasOnlyForeignParents(current))
                {
                    bool first = true;
                    builder.Append(Crop(current.ID) + " : ");
                    for (IDictionaryEnumerator f = current.Parents.GetEnumerator(); f.MoveNext(); ) {
                        if (first) {
                            builder.Append(((OwlClass)f.Value).ID);
                            first = false;
                        } else {
                            builder.Append(", " + ((OwlClass)f.Value).ID);
                        }
                    }
                    builder.Append("\n");
                    AppendChildren(builder, Classes, current, 1);
                }
            }

            builder.Append("== Roles: " + Roles.Count + "\n");
            for (IDictionaryEnumerator e = Roles.GetEnumerator(); e.MoveNext(); )
            {
                if (((OwlRole)e.Value).Parents.Count == 0)
                {
                    OwlRole role = ((OwlRole)e.Value);
                    builder.Append(Crop(role.ID) + ": " + Crop(role.Domain) + " -> " + Crop(role.Range) + "\n");
                    AppendChildren(builder, Roles, ((OwlRole)e.Value), 1);
                }
            }

            builder.Append("=== <<< Ontology: " + Uri + " ===");
            
            return builder.ToString();
        }
        
        private bool HasOnlyForeignParents(OwlClass node) {
            for (IDictionaryEnumerator e = node.Parents.GetEnumerator(); e.MoveNext(); ) {
                if (((OwlClass)e.Value).ID.IndexOf(_Uri) != -1)
                    return false;
            }
            
            return true;
        }

        private void AppendChildren(StringBuilder builder, OwlClassDictionary all, OwlClass parent, int level)
        {
            for (IDictionaryEnumerator e = all.GetEnumerator(); e.MoveNext(); )
            {
                if (((OwlClass)e.Value).Parents.Contains(parent.ID))
                {
                    for (int i = 0; i < level; i++)
                    {
                        builder.Append("  ");
                    }
                    builder.Append(Crop(((OwlClass)e.Value).ID) + "\n");
                    AppendChildren(builder, all, ((OwlClass)e.Value), level + 1);
                }
            }
        }

        private void AppendChildren(StringBuilder builder, OwlRoleDictionary all, OwlRole parent, int level)
        {
            for (IDictionaryEnumerator e = all.GetEnumerator(); e.MoveNext(); )
            {
                if (((OwlRole)e.Value).Parents.Contains(parent.ID))
                {
                    for (int i = 0; i < level; i++)
                    {
                        builder.Append("  ");
                    }
                    OwlRole role = ((OwlRole)e.Value);
                    builder.Append(Crop(role.ID) + ": " + Crop(role.Range) + " -> " + Crop(role.Domain) + "\n");
                    AppendChildren(builder, all, ((OwlRole)e.Value), level + 1);
                }
            }
        }
    }
}
