/*

  ORe - OWL ontology relevance library for Microsoft .NET
  (http://ore.mailabs.ru)

  Copyright (c) 2005, 2006 Sergey V. Mikhanov (MAILabs)

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  Version: $Id: $
  
*/

using System;
using System.Text;

namespace Mailabs.Reasoning
{
    /// <summary>
    /// ,       .   ,  
    ///      --          ID,
    /// . .        ,   --
    ///   ,  ,  ,   .
    /// 
    /// :
    /// 1. : InterestedIn;
    ///    : SergeyMikhanov = ForAll InterestedIn AI
    ///         ,     "-",   --
    ///     ,   -  (   ,      ForAll).
    /// 2. : Likes;
    ///    : DenisZarin = Exists Likes GrishkovetsBook
    ///         ,     "",   --
    ///     ,    - (     ,   ,
    ///          Exists).
    /// </summary>
    [Serializable]
    public abstract class OwlRestriction
    {
        public readonly int Descriminator;

        public readonly OwlRole Role;
        public readonly OwlClass Parent;
        public readonly OwlRestrictionType Type;
        public readonly bool Negated;

        private DateTime validFrom = DateTime.MinValue;
        private DateTime validTo = DateTime.MaxValue;

        public OwlRestriction(OwlRole role, OwlClass parent, OwlRestrictionType type, bool negated, int descriminator)
        {
            Role = role;
            Parent = parent;
            Type = type;
            Negated = negated;
            Descriminator = descriminator;
        }

        public virtual OwlClass GetChildClass()
        {
            throw new InvalidDescriminatorException("The child element of this role is not OwlClass, check Descriminator of role");
        }

        public virtual object GetChildValue()
        {
            throw new InvalidDescriminatorException("The child element of this role is not Object, check Descriminator of role");
        }

        public override bool Equals(object o)
        {
            if (o == null)
                return false;

            if (!(o is OwlRestriction))
                return false;

            OwlRestriction res = (OwlRestriction)o;

            switch (Descriminator)
            {
                case OWL_CLASS_RESTRICTION:
                    return res.Parent.ID == this.Parent.ID &&
                        res.GetChildClass().ID == this.GetChildClass().ID &&
                        res.Role.ID == this.Role.ID &&
                        res.Type == this.Type;
                case PRIMITIVE_VALUE_RESTRICTION:
                    return res.Parent.ID == this.Parent.ID &&
                        res.GetChildValue().Equals(this.GetChildValue()) &&
                        res.Role.ID == this.Role.ID &&
                        res.Type == this.Type;
                default:
                    return false;
            }
        }

        public void SetValidityPeriod(DateTime from, DateTime to)
        {
            validFrom = from;
            validTo = to;
        }

        public void SetValidityPeriod(DateTime from, TimeSpan length)
        {
            validFrom = from;
            validTo = new DateTime(from.Ticks);
            validTo.Add(length);
        }

        public bool IsValid(DateTime moment)
        {
            if (moment < validFrom || validTo < moment)
                return false;

            return true;
        }

        public const int OWL_CLASS_RESTRICTION = 0;
        public const int PRIMITIVE_VALUE_RESTRICTION = 1;
    }

    [Serializable]
    public class OwlClassRestriction : OwlRestriction
    {
        private OwlClass Child;

        public OwlClassRestriction(OwlRole role, OwlClass parent, OwlClass child, OwlRestrictionType type)
            : this(role, parent, child, type, false) { }

        public OwlClassRestriction(OwlRole role, OwlClass parent, OwlClass child, OwlRestrictionType type, bool negated)
            : base(role, parent, type, negated, OWL_CLASS_RESTRICTION)
        {
            Child = child;
        }

        public override OwlClass GetChildClass()
        {
            return Child;
        }

        public override string ToString()
        {
            return "OWL_CLASS_RESTRICTION: {" + Parent.RelativeID + " -(" + Role.RelativeID + ")-> " + Child.RelativeID + "}";
        }
    }

    [Serializable]
    public class PrimitiveValueRestriction : OwlRestriction
    {
        private object Child;

        public PrimitiveValueRestriction(OwlRole role, OwlClass parent, object child, OwlRestrictionType type)
            : this(role, parent, child, type, false) { }

        public PrimitiveValueRestriction(OwlRole role, OwlClass parent, object child, OwlRestrictionType type, bool negated)
            : base(role, parent, type, negated, PRIMITIVE_VALUE_RESTRICTION)
        {
            Child = child;
        }

        public override object GetChildValue()
        {
            return Child;
        }

        public override string ToString()
        {
            return "PRIMITIVE_VALUE_RESTRICTION: {" + Parent.RelativeID + " -(" + Role.RelativeID + ")-> " + Child + "}";
        }
    }

    public enum OwlRestrictionType
    {
        ForAll, Exists, MaxCardinality, MinCardinality
    }

    public class InvalidDescriminatorException : Exception
    {
        public InvalidDescriminatorException(string message) : base(message) {}
    }
}
