/*

  ORe - OWL ontology relevance library for Microsoft .NET
  (http://ore.mailabs.ru)

  Copyright (c) 2005, 2006 Sergey V. Mikhanov (MAILabs)

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  Version: $Id: $
  
*/

using System;
using System.Collections;

using Drive.Rdf;

using log4net;

namespace Mailabs.Reasoning
{
    /// <summary>
    /// ,      OWL/RDF.
    /// </summary>
    internal class RdfHelper
    {
        private static readonly ILog logger = LogManager.GetLogger(typeof(RdfHelper));
    
        private RdfHelper() {}

        public static void Load(Ontology ontology, string ontologyUri, LoadCallback callback) 
        {
            logger.Debug("RDF Helper initialized");
            callback.OnHelperInit();

            IRdfParserFactory factory = new RdfParserFactory();
            IRdfParser parser = factory.GetRdfXmlParser();

            logger.Debug("Starting parse process...");
            callback.OnBeforeParse();

            IRdfGraph rdfGraph = parser.ParseRdf(ontologyUri);

            logger.Debug("Parse done.");
            callback.OnAfterParse(rdfGraph.Nodes.Count);

            ontology._Uri = rdfGraph.NameSpaces[Consts.XMLNS];

            int nodeNum = 0;

            for (IEnumerator e = rdfGraph.Nodes.GetEnumerator(); e.MoveNext(); ) 
            {
                nodeNum++;
                logger.Debug("Loading node " + nodeNum + " of " + rdfGraph.Nodes.Count);
                callback.OnNodeLoad(nodeNum);

                IRdfNode node = (IRdfNode)((DictionaryEntry)e.Current).Value;
                
                if (RdfHelper.IsSystem(node))
                    continue;

                if (node.ChildEdges.Count == 0)
                    continue;

                if (node.ChildEdges[RdfConsts.RDF_TYPE] == null || node.ChildEdges[RdfConsts.RDF_TYPE].Count == 0 || node.ChildEdges[RdfConsts.RDF_TYPE].Count > 1) 
                    continue;

                string type = node.ChildEdges[RdfConsts.RDF_TYPE, 0].ChildNode.ID;
                
                switch (type) 
                {
                    case RdfConsts.OWL_ONTOLOGY: 
                    {
                        break;
                    }
                    case RdfConsts.OWL_CLASS: 
                    case RdfConsts.RDFS_CLASS: 
                    {
                        OwlClass owlClass = ontology.Classes[node.ID];
                        if (owlClass == null) 
                        {
                            if (node.ID.StartsWith(ontology.Uri)) 
                            {
                                // this class is not remote
                                owlClass = new OwlClass(node.ID);
                                ontology.Classes.Add(owlClass.ID, owlClass);
                            }
                        }

                        ProcessEdges(node.ChildEdges[RdfConsts.RDFS_SUBCLASS_OF], owlClass, owlClass.Parents, owlClass.Restrictions, ontology, false);
                        ProcessEdges(node.ChildEdges[RdfConsts.OWL_EQUIVALENT_CLASS], owlClass, owlClass.Equivalents, owlClass.Restrictions, ontology, false);
                        ProcessEdges(node.ChildEdges[RdfConsts.OWL_COMPLEMENT_CLASS], owlClass, owlClass.Complements, owlClass.Restrictions, ontology, true);
                        break;
                    }
                    case RdfConsts.OWL_OBJECT_PROPERTY:
                    {
                        OwlRole role = ontology.Roles[node.ID];
                        if (role == null) 
                        {
                            if (node.ID.StartsWith(ontology.Uri))
                            {
                                // this role is not remote
                                string range = node.ChildEdges[RdfConsts.RDFS_RANGE] != null ? node.ChildEdges[RdfConsts.RDFS_RANGE, 0].ChildNode.ID : RdfConsts.OWL_THING;
                                string domain = node.ChildEdges[RdfConsts.RDFS_DOMAIN] != null ? node.ChildEdges[RdfConsts.RDFS_DOMAIN, 0].ChildNode.ID : RdfConsts.OWL_THING;

                                role = new OwlRole(node.ID, range, domain);
                                ontology.Roles.Add(role.ID, role);
                            }
                        }

                        IRdfEdgeList parents = node.ChildEdges[RdfConsts.RDFS_SUBPROPERTY_OF];

                        if (parents != null) 
                        {
                            for (int i = 0; i < parents.Count; i++) 
                            {
                                OwlRole parentRole = ontology.Roles[parents[i].ChildNode.ID];
                                if (parentRole == null) 
                                {
                                    if (parents[i].ChildNode.ID.StartsWith(ontology.Uri))
                                    {
                                        // this role is not remote
                                        string range = parents[i].ChildNode.ChildEdges[RdfConsts.RDFS_RANGE] != null ? parents[i].ChildNode.ChildEdges[RdfConsts.RDFS_RANGE, 0].ChildNode.ID : RdfConsts.OWL_THING;
                                        string domain = parents[i].ChildNode.ChildEdges[RdfConsts.RDFS_DOMAIN] != null ? parents[i].ChildNode.ChildEdges[RdfConsts.RDFS_DOMAIN, 0].ChildNode.ID : RdfConsts.OWL_THING;

                                        parentRole = new OwlRole(parents[i].ChildNode.ID, range, domain);
                                        ontology.Roles.Add(parentRole.ID, parentRole);
                                    } 
                                    else 
                                    {
                                        // this role is remote
                                        parentRole = new OwlRole(parents[i].ChildNode.ID, true);
                                    }
                                }

                                role.Parents.Add(parentRole.ID, parentRole);
                            }
                        }
                        break;
                    }
                }
            }
        }

        private static void ProcessEdges(IRdfEdgeList edges, OwlClass owner, OwlClassDictionary classes, OwlRestrictionCollection restrictions, Ontology ontology, bool negate) 
        {
            if (edges == null)
                return;

            for (int i = 0; i < edges.Count; i++) 
            {
                // this is the explicit definition of parent/equivalent/complement class
                if (!IsSystem(edges[i].ChildNode) && !IsAnonymous(edges[i].ChildNode)) 
                {
                    OwlClass parent;
                    if (edges[i].ChildNode.ID.StartsWith(ontology.Uri)) 
                    {
                        // this class is not remote
                        parent = new OwlClass(edges[i].ChildNode.ID);
                        classes.Add(parent.ID, parent);

                        if (!ontology.Classes.Contains(parent.ID)) 
                        {
                            ontology.Classes.Add(parent.ID, parent);
                        }
                    } 
                    else 
                    {
                        // this class is remote
                        parent = new OwlClass(edges[i].ChildNode.ID, true);
                        classes.Add(parent.ID, parent);
                    }
                } 
                            
                // this is the definition of parent/equivalent/complement class using anonymous restriction
                if (IsAnonymous(edges[i].ChildNode))
                {
                    if (edges[i].ChildNode.ChildEdges[RdfConsts.RDF_TYPE, 0].ChildNode.ID == RdfConsts.OWL_RESTRICTION) 
                    {
                        IRdfNode restrNode = edges[i].ChildNode;
                        IRdfNode anonNode = restrNode.ChildEdges[RdfConsts.OWL_ON_PROPERTY, 0].ChildNode;
                        
                        string roleId = anonNode.ID;
                        OwlRole role = ontology.Roles[roleId];
                        if (role == null) 
                        {
                            if (roleId.StartsWith(ontology.Uri)) 
                            {
                                // this role is not remote
                                string range = anonNode.ChildEdges[RdfConsts.RDFS_RANGE, 0].ChildNode.ID;
                                string domain = anonNode.ChildEdges[RdfConsts.RDFS_DOMAIN, 0].ChildNode.ID;

                                role = new OwlRole(roleId, range, domain);
                                ontology.Roles.Add(role.ID, role);
                            } 
                            else 
                            {
                                role = new OwlRole(roleId, true);
                            }
                        }

                        OwlRestrictionType type;

                        if (restrNode.ChildEdges[RdfConsts.OWL_ALL_VALUES_FROM] != null) 
                        {
                            type = OwlRestrictionType.ForAll;
                        } 
                        else if (restrNode.ChildEdges[RdfConsts.OWL_SOME_VALUES_FROM] != null) 
                        {
                            type = OwlRestrictionType.Exists;
                        }
                        else if (restrNode.ChildEdges[RdfConsts.OWL_MIN_CARDINALITY] != null) 
                        {
                            type = OwlRestrictionType.MinCardinality;
                        }
                        else if (restrNode.ChildEdges[RdfConsts.OWL_MAX_CARDINALITY] != null) 
                        {
                            type = OwlRestrictionType.MaxCardinality;
                        } 
                        else 
                        {
                            throw new OwlParseException("Restriction type is not defined for class '" + owner.ID + "'");
                        }

                        string objClassId;

                        switch (type) 
                        {
                            case OwlRestrictionType.ForAll:
                                objClassId = restrNode.ChildEdges[RdfConsts.OWL_ALL_VALUES_FROM, 0].ChildNode.ID;
                                break;
                            case OwlRestrictionType.Exists:
                                objClassId = restrNode.ChildEdges[RdfConsts.OWL_SOME_VALUES_FROM, 0].ChildNode.ID;
                                break;
                            case OwlRestrictionType.MaxCardinality:
                            case OwlRestrictionType.MinCardinality:
                                // TODO: add cardinality support here
                                // NOTE: in Maorie project it is not supported
                                continue;
                            default:
                                throw new OwlParseException("Restriction type is not defined for class '" + owner.ID + "'");
                        }

                        OwlClass objClass = ontology.Classes[objClassId];
                        if (objClass == null) 
                        {
                            if (objClassId.StartsWith(ontology.Uri)) 
                            {
                                // this class is not remote
                                objClass = new OwlClass(objClassId);
                                ontology.Classes.Add(objClass.ID, objClass);
                            } 
                            else 
                            {
                                // this class is remote
                                objClass = new OwlClass(objClassId, true);
                            }
                        }

                        OwlRestriction restriction = new OwlClassRestriction(role, owner, objClass, type, negate);

                        restrictions.Add(restriction);
                    }
                }
            }
        }

        private static bool IsSystem(IRdfNode node)
        {
            return node.ID.StartsWith(Namespaces.RDF) ||
                node.ID.StartsWith(Namespaces.RDFS) ||
                node.ID.StartsWith(Namespaces.OWL) ||
                node.ID.StartsWith(Namespaces.XMLS);
        }

        private static bool IsAnonymous(IRdfNode node) 
        {
            return node.ID.StartsWith(Consts.BLANK_ID);
        }
    }
}
